@@ -81,13 +81,14 @@ cli_alloc_out(cli *c, int size)
 	o = c->tx_buf;
       else
 	{
-	  o = mb_alloc(c->pool, sizeof(struct cli_out) + CLI_TX_BUF_SIZE);
+	  o = alloc_page();
+	  c->tx_pending_count++;
 	  if (c->tx_write)
 	    c->tx_write->next = o;
 	  else
 	    c->tx_buf = o;
 	  o->wpos = o->outpos = o->buf;
-	  o->end = o->buf + CLI_TX_BUF_SIZE;
+	  o->end = (void *) o + page_size;
 	}
       c->tx_write = o;
       if (!c->tx_pos)
@@ -157,43 +158,6 @@ cli_vprintf(cli *c, int code, const char *msg, va_list args)
   memcpy(cli_alloc_out(c, size), buf, size);
 }
 
-static void
-cli_copy_message(cli *c)
-{
-  byte *p, *q;
-  uint cnt = 2;
-
-  if (c->ring_overflow)
-    {
-      byte buf[64];
-      int n = bsprintf(buf, "<%d messages lost>\n", c->ring_overflow);
-      c->ring_overflow = 0;
-      memcpy(cli_alloc_out(c, n), buf, n);
-    }
-  p = c->ring_read;
-  while (*p)
-    {
-      cnt++;
-      p++;
-      if (p == c->ring_end)
-	p = c->ring_buf;
-      ASSERT(p != c->ring_write);
-    }
-  c->async_msg_size += cnt;
-  q = cli_alloc_out(c, cnt);
-  *q++ = '+';
-  p = c->ring_read;
-  do
-    {
-      *q = *p++;
-      if (p == c->ring_end)
-	p = c->ring_buf;
-    }
-  while (*q++);
-  c->ring_read = p;
-  q[-1] = '\n';
-}
-
 static void
 cli_hello(cli *c)
 {
@@ -204,19 +168,18 @@ cli_hello(cli *c)
 static void
 cli_free_out(cli *c)
 {
-  struct cli_out *o, *p;
+  for (struct cli_out *o = c->tx_buf, *n; o; o = n)
+  {
+    n = o->next;
+    free_page(o);
+    c->tx_pending_count--;
+  }
 
-  if (o = c->tx_buf)
-    {
-      o->wpos = o->outpos = o->buf;
-      while (p = o->next)
-	{
-	  o->next = p->next;
-	  mb_free(p);
-	}
-    }
+  c->tx_buf = NULL;
   c->tx_write = c->tx_pos = NULL;
   c->async_msg_size = 0;
+
+  ASSERT_DIE(c->tx_pending_count == 0);
 }
 
 void
@@ -226,6 +189,38 @@ cli_written(cli *c)
   ev_schedule(c->event);
 }
 
+/* A dummy resource to show and free memory pages allocated for pending TX */
+struct cli_tx_resource {
+  resource r;
+  struct cli *c;
+};
+
+static void
+cli_tx_resource_free(resource *r)
+{
+  cli_free_out(SKIP_BACK(struct cli_tx_resource, r, r)->c);
+}
+
+static void
+cli_tx_resource_dump(struct dump_request *dreq UNUSED, resource *r UNUSED) {}
+
+static struct resmem
+cli_tx_resource_memsize(resource *r)
+{
+  return (struct resmem) {
+    .effective = SKIP_BACK(struct cli_tx_resource, r, r)->c->tx_pending_count * page_size,
+    .overhead = sizeof(struct cli_tx_resource),
+  };
+}
+
+static struct resclass cli_tx_resource_class = {
+  .name = "CLI TX buffers",
+  .size = sizeof (struct cli_tx_resource),
+  .free = cli_tx_resource_free,
+  .dump = cli_tx_resource_dump,
+  .memsize = cli_tx_resource_memsize,
+};
+
 
 static byte *cli_rh_pos;
 static uint cli_rh_len;
@@ -255,22 +250,26 @@ cli_command(struct cli *c)
   struct config f;
   int res;
 
-  if (config->cli_debug > 1)
+  if (OBSREF_GET(config)->cli_debug > 1)
     log(L_TRACE "CLI: %s", c->rx_buf);
   bzero(&f, sizeof(f));
   f.mem = c->parser_pool;
-  f.pool = rp_new(c->pool, "Config");
+  f.pool = rp_new(c->pool, the_bird_domain.the_bird, "Config");
+  obstacle_target_init(&f.obstacles, &f.obstacles_cleared, c->pool, "Config");
+
   init_list(&f.symbols);
   cf_read_hook = cli_cmd_read_hook;
   cli_rh_pos = c->rx_buf;
   cli_rh_len = strlen(c->rx_buf);
   cli_rh_trick_flag = 0;
   this_cli = c;
+  this_cli->main_config = OBSREF_GET(config);
   lp_flush(c->parser_pool);
-  res = cli_parse(&f);
+  res = cli_parse(this_cli->main_config, &f);
   if (!res)
     cli_printf(c, 9001, f.err_msg);
 
+  this_cli->main_config = NULL;
   config_free(&f);
 }
 
@@ -280,10 +279,6 @@ cli_event(void *data)
   cli *c = data;
   int err;
 
-  while (c->ring_read != c->ring_write &&
-      c->async_msg_size < CLI_MAX_ASYNC_QUEUE)
-    cli_copy_message(c);
-
   if (c->tx_pos)
     ;
   else if (c->cont)
@@ -299,24 +294,27 @@ cli_event(void *data)
 	cli_command(c);
     }
 
-  cli_write_trigger(c);
+  if (c->tx_pos)
+    cli_write_trigger(c);
+  else
+    ev_schedule(c->event);
 }
 
 cli *
-cli_new(void *priv, struct cli_config *cf)
+cli_new(struct birdsock *sock, struct cli_config *cf)
 {
-  pool *p = rp_new(cli_pool, "CLI");
-  cli *c = mb_alloc(p, sizeof(cli));
+  pool *p = rp_new(cli_pool, the_bird_domain.the_bird, "CLI");
+  struct cli_tx_resource *ctr = ralloc(p, &cli_tx_resource_class);
+  cli *c = ctr->c = mb_alloc(p, sizeof(cli));
 
   bzero(c, sizeof(cli));
   c->pool = p;
-  c->priv = priv;
+  c->sock = sock;
   c->event = ev_new(p);
   c->event->hook = cli_event;
   c->event->data = c;
   c->cont = cli_hello;
   c->parser_pool = lp_new_default(c->pool);
-  c->show_pool = lp_new_default(c->pool);
   c->rx_buf = mb_alloc(c->pool, CLI_RX_BUF_SIZE);
 
   if (cf->restricted)
@@ -336,73 +334,6 @@ cli_kick(cli *c)
 static list cli_log_hooks;
 static int cli_log_inited;
 
-void
-cli_set_log_echo(cli *c, uint mask, uint size)
-{
-  if (c->ring_buf)
-    {
-      mb_free(c->ring_buf);
-      c->ring_buf = c->ring_end = c->ring_read = c->ring_write = NULL;
-      rem_node(&c->n);
-    }
-  c->log_mask = mask;
-  if (mask && size)
-    {
-      c->ring_buf = mb_alloc(c->pool, size);
-      c->ring_end = c->ring_buf + size;
-      c->ring_read = c->ring_write = c->ring_buf;
-      add_tail(&cli_log_hooks, &c->n);
-      c->log_threshold = size / 8;
-    }
-  c->ring_overflow = 0;
-}
-
-void
-cli_echo(uint class, byte *msg)
-{
-  unsigned len, free, i, l;
-  cli *c;
-  byte *m;
-
-  if (!cli_log_inited || EMPTY_LIST(cli_log_hooks))
-    return;
-  len = strlen(msg) + 1;
-  WALK_LIST(c, cli_log_hooks)
-    {
-      if (!(c->log_mask & (1 << class)))
-	continue;
-      if (c->ring_read <= c->ring_write)
-	free = (c->ring_end - c->ring_buf) - (c->ring_write - c->ring_read + 1);
-      else
-	free = c->ring_read - c->ring_write - 1;
-      if ((len > free) ||
-	  (free < c->log_threshold && class < (unsigned) L_INFO[0]))
-	{
-	  c->ring_overflow++;
-	  continue;
-	}
-      if (c->ring_read == c->ring_write)
-	ev_schedule(c->event);
-      m = msg;
-      l = len;
-      while (l)
-	{
-	  if (c->ring_read <= c->ring_write)
-	    i = c->ring_end - c->ring_write;
-	  else
-	    i = c->ring_read - c->ring_write;
-	  if (i > l)
-	    i = l;
-	  memcpy(c->ring_write, m, i);
-	  m += i;
-	  l -= i;
-	  c->ring_write += i;
-	  if (c->ring_write == c->ring_end)
-	    c->ring_write = c->ring_buf;
-	}
-    }
-}
-
 /* Set time format override for the current session */
 void
 cli_set_timeformat(cli *c, const struct timeformat tf)
@@ -439,12 +370,18 @@ extern cli *cmd_reconfig_stored_cli;
 void
 cli_free(cli *c)
 {
-  cli_set_log_echo(c, 0, 0);
-  if (c->cleanup)
-    c->cleanup(c);
+  bool done = c->cleanup ? c->cleanup(c) : true;
+
   if (c == cmd_reconfig_stored_cli)
     cmd_reconfig_stored_cli = NULL;
-  rfree(c->pool);
+
+  if (done)
+    rp_free(c->pool);
+  else
+  {
+    sk_close(c->sock);
+    c->sock = NULL;
+  }
 }
 
 /**
@@ -456,7 +393,7 @@ cli_free(cli *c)
 void
 cli_init(void)
 {
-  cli_pool = rp_new(&root_pool, "CLI");
+  cli_pool = rp_new(&root_pool, the_bird_domain.the_bird, "CLI");
   init_list(&cli_log_hooks);
   cli_log_inited = 1;
 }
