@@ -23,7 +23,27 @@
 #include "filter/filter.h"
 #include "filter/f-inst.h"
 
+/* Pool for protocol persistent data set by protocols themselves */
+struct proto_common_private {
+  DOMAIN(attrs) lock;
+  struct proto_common_private **locked_at;
+
+  pool *pool;
+};
+
+typedef union proto_common {
+  DOMAIN(attrs) lock;
+  struct proto_common_private priv;
+} proto_common;
+static proto_common proto_common_pub;
+
+LOBJ_UNLOCK_CLEANUP(proto_common, attrs);
+#define PROTO_COMMON_LOCK(pcm)	LOBJ_LOCK(&proto_common_pub, pcm, proto_common, attrs);
+
+/* Top pool for protocol data */
 pool *proto_pool;
+
+
 static TLIST_LIST(proto) global_proto_list;
 
 static list STATIC_LIST_INIT(protocol_list);
@@ -1912,7 +1932,10 @@ proto_rethink_goal(struct proto *p)
 
     proto_rem_node(&global_proto_list, p);
     rfree(p->event);
-    mb_free(p->message);
+    {
+      PROTO_COMMON_LOCK(pcm);
+      mb_free(p->message);
+    }
     mb_free(p);
     if (!nc)
       goto done;
@@ -1938,6 +1961,18 @@ done:
 	);
 }
 
+struct proto_rethink_goal_deferred {
+  struct deferred_call dc;
+  struct proto *p;
+};
+
+static void
+proto_rethink_goal_deferred(struct deferred_call *dc)
+{
+  SKIP_BACK_DECLARE(struct proto_rethink_goal_deferred, prgd, dc, dc);
+  proto_rethink_goal(prgd->p);
+}
+
 struct proto *
 proto_spawn(struct proto_config *cf, uint disabled)
 {
@@ -1963,7 +1998,11 @@ proto_enable(struct proto *p)
   ASSERT_DIE(birdloop_inside(&main_birdloop));
   bool changed = p->disabled;
   p->disabled = 0;
-  proto_rethink_goal(p);
+  struct proto_rethink_goal_deferred prgd = {
+    .dc.hook = proto_rethink_goal_deferred,
+    .p = p,
+  };
+  defer_call(&prgd.dc, sizeof prgd);
   return changed;
 }
 
@@ -2243,6 +2282,11 @@ void
 protos_build(void)
 {
   proto_pool = rp_new(&root_pool, the_bird_domain.the_bird, "Protocols");
+  proto_common_pub.lock = DOMAIN_NEW(attrs);
+  {
+    PROTO_COMMON_LOCK(pcm);
+    pcm->pool = rp_new(proto_pool, pcm->lock.attrs, "Common persistent protocol data");
+  }
 
   /* Protocol attributes */
   ea_register_init(&ea_name);
@@ -2353,6 +2397,8 @@ proto_schedule_down(struct proto *p, byte restart, byte code)
 void
 proto_set_message(struct proto *p, char *msg, int len)
 {
+  PROTO_COMMON_LOCK(pcm);
+
   mb_free(p->message);
   p->message = NULL;
 
@@ -2365,7 +2411,7 @@ proto_set_message(struct proto *p, char *msg, int len)
   if (!len)
     return;
 
-  p->message = mb_alloc(proto_pool, len + 1);
+  p->message = mb_alloc(pcm->pool, len + 1);
   memcpy(p->message, msg, len);
   p->message[len] = 0;
 }
@@ -2826,7 +2872,11 @@ proto_cmd_enable(struct proto *p, uintptr_t arg, int cnt UNUSED)
   log(L_INFO "Enabling protocol %s", p->name);
   p->disabled = 0;
   proto_set_message(p, (char *) arg, -1);
-  proto_rethink_goal(p);
+  struct proto_rethink_goal_deferred prgd = {
+    .dc.hook = proto_rethink_goal_deferred,
+    .p = p,
+  };
+  defer_call(&prgd.dc, sizeof prgd);
   cli_msg(-11, "%s: enabled", p->name);
 }
 
