@@ -252,6 +252,7 @@ struct ospf_proto
   u32 last_vlink_id;		/* Interface IDs for vlinks (starts at 0x80000000) */
   struct tbf log_pkt_tbf;	/* TBF for packet messages */
   struct tbf log_lsa_tbf;	/* TBF for LSA messages */
+  ip_addr loopback_addr;	/* IP address used as common next hop (in OSPFv3-IPv4) */
 };
 
 struct ospf_area
@@ -331,6 +332,7 @@ struct ospf_iface
   struct top_hash_entry **flood_queue;	/* LSAs queued for LSUPD */
   u8 update_link_lsa;
   u8 update_net_lsa;
+  u8 loopback_addr_used;	/* The Link-LSA depends on p->loopback_addr */
   u16 flood_queue_used;		/* The current number of LSAs in flood_queue */
   u16 flood_queue_size;		/* The maximum number of LSAs in flood_queue */
   int fadj;			/* Number of fully adjacent neighbors */
@@ -939,12 +941,7 @@ struct lsadb_show_data {
   u32 router;		/* Advertising router, 0 -> all */
 };
 
-
-#define EA_OSPF_METRIC1	EA_CODE(PROTOCOL_OSPF, 0)
-#define EA_OSPF_METRIC2	EA_CODE(PROTOCOL_OSPF, 1)
-#define EA_OSPF_TAG	EA_CODE(PROTOCOL_OSPF, 2)
-#define EA_OSPF_ROUTER_ID EA_CODE(PROTOCOL_OSPF, 3)
-
+extern struct ea_class ea_ospf_metric1, ea_ospf_metric2, ea_ospf_tag, ea_ospf_router_id;
 
 /*
  * For regular networks, neighbor address must match network prefix.
@@ -1007,6 +1004,8 @@ void ospf_sh_state(struct proto *P, int verbose, int reachable);
 
 void ospf_sh_lsadb(struct lsadb_show_data *ld);
 
+extern struct rte_owner_class ospf_rte_owner_class;
+
 /* iface.c */
 void ospf_iface_chstate(struct ospf_iface *ifa, u8 state);
 void ospf_iface_sm(struct ospf_iface *ifa, int event);
