@@ -51,7 +51,7 @@ struct static_route {
   uint mpls_label;			/* Local MPLS label, -1 if unused */
   struct bfd_request *bfd_req;		/* BFD request, if BFD is used */
   union {
-    mpls_label_stack *mls;		/* MPLS label stack; may be NULL */
+    adata *mls;				/* MPLS label stack; may be NULL */
     adata *aspa;			/* ASPA provider list; may be NULL */
   };
 };
