@@ -16,8 +16,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <time.h>
+#include <sys/mman.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/uio.h>
@@ -38,6 +40,7 @@
 #include "lib/resource.h"
 #include "lib/socket.h"
 #include "lib/event.h"
+#include "lib/locking.h"
 #include "lib/timer.h"
 #include "lib/string.h"
 #include "nest/iface.h"
@@ -45,6 +48,7 @@
 #include "conf/conf.h"
 
 #include "sysdep/unix/unix.h"
+#include "sysdep/unix/io-loop.h"
 #include CONFIG_INCLUDE_SYSIO_H
 
 /* Maximum number of calls of tx handler for one socket in one
@@ -65,7 +69,15 @@
 
 struct rfile {
   resource r;
-  FILE *f;
+  struct stat stat;
+  int fd;
+  off_t limit;
+  _Atomic off_t pos;
+  void *mapping;
+};
+
+struct rfile rf_stderr = {
+  .fd = 2,
 };
 
 static void
@@ -73,7 +85,10 @@ rf_free(resource *r)
 {
   struct rfile *a = (struct rfile *) r;
 
-  fclose(a->f);
+  if (a->mapping)
+    munmap(a->mapping, a->limit);
+
+  close(a->fd);
 }
 
 static void
@@ -81,7 +96,7 @@ rf_dump(struct dump_request *dreq, resource *r)
 {
   struct rfile *a = (struct rfile *) r;
 
-  RDUMP("(FILE *%p)\n", a->f);
+  RDUMP("(fd %d)\n", a->fd);
 }
 
 static struct resclass rf_class = {
@@ -93,42 +108,183 @@ static struct resclass rf_class = {
   NULL
 };
 
-struct rfile *
-rf_open(pool *p, const char *name, const char *mode)
-{
-  FILE *f = fopen(name, mode);
-  if (!f)
-    return NULL;
-
-  struct rfile *r = ralloc(p, &rf_class);
-  r->f = f;
-
-  return r;
-}
-
-struct rfile *
-rf_fdopen(pool *p, int fd, const char *mode)
-{
-  FILE *f = fdopen(fd, mode);
-  if (!f)
-    return NULL;
-
-  struct rfile *r = ralloc(p, &rf_class);
-  r->f = f;
-
-  return r;
-}
-
-void *
-rf_file(struct rfile *f)
-{
-  return f->f;
-}
-
 int
 rf_fileno(struct rfile *f)
 {
-  return fileno(f->f);
+  return f->fd;
+}
+
+static int
+rf_open_get_fd(const char *name, enum rf_mode mode)
+{
+  int omode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+  int flags;
+
+  switch (mode)
+  {
+    case RF_APPEND:
+      flags = O_WRONLY | O_CREAT | O_APPEND;
+      break;
+
+    case RF_FIXED:
+      flags = O_RDWR | O_CREAT;
+      break;
+
+    default:
+      bug("rf_open() must have the mode set");
+  }
+
+  return open(name, flags, omode);
+}
+
+static void
+rf_stat(struct rfile *r)
+{
+  if (fstat(r->fd, &r->stat) < 0)
+    die("fstat() failed: %m");
+}
+
+struct rfile *
+rf_open(pool *p, const char *name, enum rf_mode mode, off_t limit)
+{
+  int fd = rf_open_get_fd(name, mode);
+  if (fd < 0)
+    return NULL; /* The caller takes care of printing %m. */
+
+  struct rfile *r = ralloc(p, &rf_class);
+  r->fd = fd;
+  r->limit = limit;
+
+  switch (mode)
+  {
+    case RF_APPEND:
+      rf_stat(r);
+      atomic_store_explicit(&r->pos, S_ISREG(r->stat.st_mode) ? r->stat.st_size : 0, memory_order_relaxed);
+      break;
+
+    case RF_FIXED:
+      if ((ftruncate(fd, limit) < 0)
+	  || ((r->mapping = mmap(NULL, limit, PROT_WRITE, MAP_SHARED, fd, 0)) == MAP_FAILED))
+      {
+	int erf = errno;
+	r->mapping = NULL;
+	rfree(r);
+	errno = erf;
+	return NULL;
+      }
+      break;
+
+    default:
+      bug("rf_open() must have the mode set");
+  }
+
+
+  return r;
+}
+
+off_t
+rf_size(struct rfile *r)
+{
+  return atomic_load_explicit(&r->pos, memory_order_relaxed);
+}
+
+int
+rf_same(struct rfile *a, struct rfile *b)
+{
+  rf_stat(a);
+  rf_stat(b);
+
+  return
+    (a->limit == b->limit) &&
+    (a->stat.st_mode == b->stat.st_mode) &&
+    (a->stat.st_dev == b->stat.st_dev) &&
+    (a->stat.st_ino == b->stat.st_ino);
+}
+
+void
+rf_write_crude(struct rfile *r, const char *buf, int sz)
+{
+  if (r->mapping)
+    memcpy(r->mapping, buf, sz);
+  else
+    write(r->fd, buf, sz);
+}
+
+
+int
+rf_writev(struct rfile *r, struct iovec *iov, int iov_count)
+{
+  off_t size = 0;
+  for (int i = 0; i < iov_count; i++)
+    size += iov[i].iov_len;
+
+  if (r->mapping)
+  {
+    /* Update the pointer */
+    off_t target = atomic_fetch_add_explicit(&r->pos, size, memory_order_relaxed) % r->limit;
+
+    /* Write the line */
+    for (int i = 0; i < iov_count; i++)
+    {
+      /* Take care of wrapping; this should really happen only once */
+      off_t rsz;
+      while ((rsz = r->limit - target) < (off_t) iov[i].iov_len)
+      {
+	memcpy(r->mapping + target, iov[i].iov_base, rsz);
+	iov[i].iov_base += rsz;
+	iov[i].iov_len -= rsz;
+	target = 0;
+      }
+
+      memcpy(r->mapping + target, iov[i].iov_base, iov[i].iov_len);
+      target += iov[i].iov_len;
+    }
+    return 1;
+  }
+  else if (r->limit && (atomic_fetch_add_explicit(&r->pos, size, memory_order_relaxed) + size > r->limit))
+  {
+    atomic_fetch_sub_explicit(&r->pos, size, memory_order_relaxed);
+    return 0;
+  }
+  else
+  {
+    while (size > 0)
+    {
+      /* Try to write */
+      ssize_t e = writev(r->fd, iov, iov_count);
+      if (e < 0)
+	if (errno == EINTR)
+	  continue;
+	else
+	  return 1; /* FIXME: What should we do when we suddenly can't write? */
+
+      /* It is expected that we always write the whole bunch at once */
+      if (e == size)
+	return 1;
+
+      /* Block split should not happen (we write small enough messages)
+       * but if it happens, let's try to write the rest of the log */
+      size -= e;
+      while (e > 0)
+      {
+	if ((ssize_t) iov[0].iov_len > e)
+	{
+	  /* Some bytes are remaining in the first chunk */
+	  iov[0].iov_len -= e;
+	  iov[0].iov_base += e;
+	  break;
+	}
+
+	/* First chunk written completely, get rid of it */
+	e -= iov[0].iov_len;
+	iov++;
+	iov_count--;
+	ASSERT_DIE(iov_count > 0);
+      }
+    }
+
+    return 1;
+  }
 }
 
 /*
@@ -279,12 +435,16 @@ cmd_dump_file(struct cli *cli, const char *file, const char *what, void (*dump)(
 
 btime boot_time;
 
+
 void
-times_init(struct timeloop *loop)
+times_update(void)
 {
   struct timespec ts;
   int rv;
 
+  btime old_time = current_time();
+  btime old_real_time = current_real_time();
+
   rv = clock_gettime(CLOCK_MONOTONIC, &ts);
   if (rv < 0)
     die("Monotonic clock is missing");
@@ -292,40 +452,32 @@ times_init(struct timeloop *loop)
   if ((ts.tv_sec < 0) || (((u64) ts.tv_sec) > ((u64) 1 << 40)))
     log(L_WARN "Monotonic clock is crazy");
 
-  loop->last_time = ts.tv_sec S + ts.tv_nsec NS;
-  loop->real_time = 0;
-}
-
-void
-times_update(struct timeloop *loop)
-{
-  struct timespec ts;
-  int rv;
-
-  rv = clock_gettime(CLOCK_MONOTONIC, &ts);
-  if (rv < 0)
-    die("clock_gettime: %m");
-
   btime new_time = ts.tv_sec S + ts.tv_nsec NS;
 
-  if (new_time < loop->last_time)
+  if (new_time < old_time)
     log(L_ERR "Monotonic clock is broken");
 
-  loop->last_time = new_time;
-  loop->real_time = 0;
-}
-
-void
-times_update_real_time(struct timeloop *loop)
-{
-  struct timespec ts;
-  int rv;
-
   rv = clock_gettime(CLOCK_REALTIME, &ts);
   if (rv < 0)
     die("clock_gettime: %m");
 
-  loop->real_time = ts.tv_sec S + ts.tv_nsec NS;
+  btime new_real_time = ts.tv_sec S + ts.tv_nsec NS;
+
+  if (!atomic_compare_exchange_strong_explicit(
+      &last_time,
+      &old_time,
+      new_time,
+      memory_order_acq_rel,
+      memory_order_relaxed))
+    DBG("Time update collision: last_time");
+
+  if (!atomic_compare_exchange_strong_explicit(
+      &real_time,
+      &old_real_time,
+      new_real_time,
+      memory_order_acq_rel,
+      memory_order_relaxed))
+    DBG("Time update collision: real_time");
 }
 
 btime
@@ -341,7 +493,6 @@ current_time_now(void)
   return ts.tv_sec S + ts.tv_nsec NS;
 }
 
-
 /**
  * DOC: Sockets
  *
@@ -929,11 +1080,7 @@ sk_log_error(sock *s, const char *p)
  *	Actual struct birdsock code
  */
 
-static list sock_list;
-static struct birdsock *current_sock;
-static struct birdsock *stored_sock;
-
-static inline sock *
+sock *
 sk_next(sock *s)
 {
   if (!s->n.next->next)
@@ -981,8 +1128,7 @@ sk_ssh_free(sock *s)
 
   if (ssh->channel)
   {
-    if (ssh_channel_is_open(ssh->channel))
-      ssh_channel_close(ssh->channel);
+    ssh_channel_close(ssh->channel);
     ssh_channel_free(ssh->channel);
     ssh->channel = NULL;
   }
@@ -996,6 +1142,7 @@ sk_ssh_free(sock *s)
 }
 #endif
 
+
 static void
 sk_free(resource *r)
 {
@@ -1008,20 +1155,10 @@ sk_free(resource *r)
     sk_ssh_free(s);
 #endif
 
-  if (s->fd < 0)
-    return;
+  if (s->loop)
+    birdloop_remove_socket(s->loop, s);
 
-  /* FIXME: we should call sk_stop() for SKF_THREAD sockets */
-  if (!(s->flags & SKF_THREAD))
-  {
-    if (s == current_sock)
-      current_sock = sk_next(s);
-    if (s == stored_sock)
-      stored_sock = sk_next(s);
-    rem_node(&s->n);
-  }
-
-  if (s->type != SK_SSH && s->type != SK_SSH_ACTIVE)
+  if (s->fd >= 0 && s->type != SK_SSH && s->type != SK_SSH_ACTIVE)
     close(s->fd);
 
   s->fd = -1;
@@ -1151,7 +1288,7 @@ sk_setup(sock *s)
   }
 #endif
 
-  if (s->vrf && !s->iface && (s->type != SK_TCP))
+  if (s->vrf && (s->vrf != &default_vrf) && !s->iface && (s->type != SK_TCP))
   {
     /* Bind socket to associated VRF interface.
        This is Linux-specific, but so is SO_BINDTODEVICE.
@@ -1244,12 +1381,6 @@ sk_setup(sock *s)
   return 0;
 }
 
-static void
-sk_insert(sock *s)
-{
-  add_tail(&sock_list, &s->n);
-}
-
 static int
 sk_connect(sock *s)
 {
@@ -1298,6 +1429,10 @@ sk_passive_connected(sock *s, int type)
     return 0;
   }
 
+  struct domain_generic *sock_lock = DG_IS_LOCKED(s->pool->domain) ? NULL : s->pool->domain;
+  if (sock_lock)
+    DG_LOCK(sock_lock);
+
   sock *t = sk_new(s->pool);
   t->type = type;
   t->data = s->data;
@@ -1327,13 +1462,21 @@ sk_passive_connected(sock *s, int type)
     /* FIXME: handle it better in rfree() */
     close(t->fd);
     t->fd = -1;
-    rfree(t);
-    return 1;
+    sk_close(t);
+    t = NULL;
+  }
+  else
+  {
+    birdloop_add_socket(s->loop, t);
+    sk_alloc_bufs(t);
   }
 
-  sk_insert(t);
-  sk_alloc_bufs(t);
-  s->rx_hook(t, 0);
+  if (sock_lock)
+    DG_UNLOCK(sock_lock);
+
+  if (t)
+    s->rx_hook(t, 0);
+
   return 1;
 }
 
@@ -1376,35 +1519,46 @@ sk_ssh_connect(sock *s)
     {
       int server_identity_is_ok = 1;
 
+#ifdef HAVE_SSH_OLD_SERVER_VALIDATION_API
+#define ssh_session_is_known_server	ssh_is_server_known
+#define SSH_KNOWN_HOSTS_OK		SSH_SERVER_KNOWN_OK
+#define SSH_KNOWN_HOSTS_UNKNOWN		SSH_SERVER_NOT_KNOWN
+#define SSH_KNOWN_HOSTS_CHANGED		SSH_SERVER_KNOWN_CHANGED
+#define SSH_KNOWN_HOSTS_NOT_FOUND	SSH_SERVER_FILE_NOT_FOUND
+#define SSH_KNOWN_HOSTS_ERROR		SSH_SERVER_ERROR
+#define SSH_KNOWN_HOSTS_OTHER		SSH_SERVER_FOUND_OTHER
+#endif
+
       /* Check server identity */
-      switch (ssh_is_server_known(s->ssh->session))
+      switch (ssh_session_is_known_server(s->ssh->session))
       {
 #define LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s,msg,args...) log(L_WARN "SSH Identity %s@%s:%u: " msg, (s)->ssh->username, (s)->host, (s)->dport, ## args);
-      case SSH_SERVER_KNOWN_OK:
+      case SSH_KNOWN_HOSTS_OK:
 	/* The server is known and has not changed. */
 	break;
 
-      case SSH_SERVER_NOT_KNOWN:
+      case SSH_KNOWN_HOSTS_UNKNOWN:
 	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "The server is unknown, its public key was not found in the known host file %s", s->ssh->server_hostkey_path);
+	server_identity_is_ok = 0;
 	break;
 
-      case SSH_SERVER_KNOWN_CHANGED:
+      case SSH_KNOWN_HOSTS_CHANGED:
 	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "The server key has changed. Either you are under attack or the administrator changed the key.");
 	server_identity_is_ok = 0;
 	break;
 
-      case SSH_SERVER_FILE_NOT_FOUND:
+      case SSH_KNOWN_HOSTS_NOT_FOUND:
 	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "The known host file %s does not exist", s->ssh->server_hostkey_path);
 	server_identity_is_ok = 0;
 	break;
 
-      case SSH_SERVER_ERROR:
+      case SSH_KNOWN_HOSTS_ERROR:
 	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "Some error happened");
 	server_identity_is_ok = 0;
 	break;
 
-      case SSH_SERVER_FOUND_OTHER:
-	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "The server gave use a key of a type while we had an other type recorded. " \
+      case SSH_KNOWN_HOSTS_OTHER:
+	LOG_WARN_ABOUT_SSH_SERVER_VALIDATION(s, "The server gave use a key of a type while we had another type recorded. " \
 					     "It is a possible attack.");
 	server_identity_is_ok = 0;
 	break;
@@ -1534,6 +1688,7 @@ sk_open_ssh(sock *s)
 
 /**
  * sk_open - open a socket
+ * @loop: loop
  * @s: socket
  *
  * This function takes a socket resource created by sk_new() and
@@ -1543,7 +1698,7 @@ sk_open_ssh(sock *s)
  * Result: 0 for success, -1 for an error.
  */
 int
-sk_open(sock *s)
+sk_open(sock *s, struct birdloop *loop)
 {
   int af = AF_UNSPEC;
   int fd = -1;
@@ -1703,9 +1858,7 @@ sk_open(sock *s)
     sk_alloc_bufs(s);
   }
 
-  if (!(s->flags & SKF_THREAD))
-    sk_insert(s);
-
+  birdloop_add_socket(loop, s);
   return 0;
 
 err:
@@ -1715,7 +1868,7 @@ err:
 }
 
 int
-sk_open_unix(sock *s, const char *name)
+sk_open_unix(sock *s, struct birdloop *loop, const char *name)
 {
   struct sockaddr_un sa;
   int fd;
@@ -1742,7 +1895,7 @@ sk_open_unix(sock *s, const char *name)
     return -1;
 
   s->fd = fd;
-  sk_insert(s);
+  birdloop_add_socket(loop, s);
   return 0;
 }
 
@@ -1868,6 +2021,13 @@ sk_recvmsg(sock *s)
 
 static inline void reset_tx_buffer(sock *s) { s->ttx = s->tpos = s->tbuf; }
 
+bool
+sk_tx_pending(sock *s)
+{
+  return s->ttx != s->tpos;
+}
+
+
 static int
 sk_maybe_write(sock *s)
 {
@@ -1878,7 +2038,7 @@ sk_maybe_write(sock *s)
   case SK_TCP:
   case SK_MAGIC:
   case SK_UNIX:
-    while (s->ttx != s->tpos)
+    while (sk_tx_pending(s))
     {
       e = write(s->fd, s->ttx, s->tpos - s->ttx);
 
@@ -1900,7 +2060,7 @@ sk_maybe_write(sock *s)
 
 #ifdef HAVE_LIBSSH
   case SK_SSH:
-    while (s->ttx != s->tpos)
+    while (sk_tx_pending(s))
     {
       e = ssh_channel_write(s->ssh->channel, s->ttx, s->tpos - s->ttx);
 
@@ -1983,7 +2143,12 @@ sk_send(sock *s, unsigned len)
 {
   s->ttx = s->tbuf;
   s->tpos = s->tbuf + len;
-  return sk_maybe_write(s);
+
+  int e = sk_maybe_write(s);
+  if (e == 0) /* Trigger thread poll reload to poll this socket's write. */
+    socket_changed(s);
+
+  return e;
 }
 
 /**
@@ -2030,7 +2195,7 @@ call_rx_hook(sock *s, int size)
   if (s->rx_hook(s, size))
   {
     /* We need to be careful since the socket could have been deleted by the hook */
-    if (current_sock == s)
+    if (s->loop->sock_active == s)
       s->rpos = s->rbuf;
   }
 }
@@ -2191,7 +2356,7 @@ sk_write_noflush(sock *s)
 #endif
 
   default:
-    if (s->ttx != s->tpos && sk_maybe_write(s) > 0)
+    if (sk_tx_pending(s) && sk_maybe_write(s) > 0)
     {
       if (s->tx_hook)
 	s->tx_hook(s);
@@ -2237,12 +2402,14 @@ sk_dump_all(struct dump_request *dreq)
   sock *s;
 
   RDUMP("Open sockets:\n");
-  WALK_LIST(n, sock_list)
+  dreq->indent += 3;
+  WALK_LIST(n, main_birdloop.sock_list)
   {
     s = SKIP_BACK(sock, n, n);
     RDUMP("%p ", s);
     sk_dump(dreq, &s->r);
   }
+  dreq->indent -= 3;
   RDUMP("\n");
 }
 
@@ -2264,32 +2431,20 @@ struct event_log_entry
 static struct event_log_entry event_log[EVENT_LOG_LENGTH];
 static struct event_log_entry *event_open;
 static int event_log_pos, event_log_num, watchdog_active;
-static btime last_time;
+static btime last_io_time;
 static btime loop_time;
 
 static void
 io_update_time(void)
 {
-  struct timespec ts;
-  int rv;
-
-  /*
-   * This is third time-tracking procedure (after update_times() above and
-   * times_update() in BFD), dedicated to internal event log and latency
-   * tracking. Hopefully, we consolidate these sometimes.
-   */
-
-  rv = clock_gettime(CLOCK_MONOTONIC, &ts);
-  if (rv < 0)
-    die("clock_gettime: %m");
-
-  last_time = ts.tv_sec S + ts.tv_nsec NS;
+  last_io_time = current_time();
 
   if (event_open)
   {
-    event_open->duration = last_time - event_open->timestamp;
+    event_open->duration = last_io_time - event_open->timestamp;
 
-    if (event_open->duration > config->latency_limit)
+    struct global_runtime *gr = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+    if (event_open->duration > gr->latency_limit)
       log(L_WARN "Event 0x%p 0x%p took %u.%03u ms",
 	  event_open->hook, event_open->data, (uint) (event_open->duration TO_MS), (uint) (event_open->duration % 1000));
 
@@ -2307,23 +2462,24 @@ io_update_time(void)
  * entry is kept open (in @event_open) so the duration can be filled later.
  */
 void
-io_log_event(void *hook, void *data)
+io_log_event(void *hook, void *data, uint flag)
 {
-  if (config->latency_debug)
+  struct global_runtime *gr = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+  if (gr->latency_debug & flag)
     io_update_time();
 
   struct event_log_entry *en = event_log + event_log_pos;
 
   en->hook = hook;
   en->data = data;
-  en->timestamp = last_time;
+  en->timestamp = last_io_time;
   en->duration = 0;
 
   event_log_num++;
   event_log_pos++;
   event_log_pos %= EVENT_LOG_LENGTH;
 
-  event_open = config->latency_debug ? en : NULL;
+  event_open = (gr->latency_debug & flag) ? en : NULL;
 }
 
 static inline void
@@ -2338,21 +2494,22 @@ io_log_dump(struct dump_request *dreq)
 {
   int i;
 
-  RDUMP("Event log:");
+  RDUMP("Event log:\n");
   for (i = 0; i < EVENT_LOG_LENGTH; i++)
   {
     struct event_log_entry *en = event_log + (event_log_pos + i) % EVENT_LOG_LENGTH;
     if (en->hook)
-      RDUMP("  Event 0x%p 0x%p at %8d for %d ms", en->hook, en->data,
-	  (int) ((last_time - en->timestamp) TO_MS), (int) (en->duration TO_MS));
+      RDUMP("  Event 0x%p 0x%p at %8d for %d ms\n", en->hook, en->data,
+	  (int) ((last_io_time - en->timestamp) TO_MS), (int) (en->duration TO_MS));
   }
 }
 
 void
 watchdog_sigalrm(int sig UNUSED)
 {
-  /* Update last_time and duration, but skip latency check */
-  config->latency_limit = 0xffffffff;
+  /* Update last_io_time and duration, but skip latency check */
+  struct global_runtime *gr = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+  gr->latency_limit = 0xffffffff;
   io_update_time();
 
   debug_safe("Watchdog timer timed out\n");
@@ -2366,7 +2523,7 @@ watchdog_start1(void)
 {
   io_update_time();
 
-  loop_time = last_time;
+  loop_time = last_io_time;
 }
 
 static inline void
@@ -2374,12 +2531,13 @@ watchdog_start(void)
 {
   io_update_time();
 
-  loop_time = last_time;
+  loop_time = last_io_time;
   event_log_num = 0;
 
-  if (config->watchdog_timeout)
+  struct global_runtime *gr = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+  if (gr->watchdog_timeout)
   {
-    alarm(config->watchdog_timeout);
+    alarm(gr->watchdog_timeout);
     watchdog_active = 1;
   }
 }
@@ -2395,8 +2553,9 @@ watchdog_stop(void)
     watchdog_active = 0;
   }
 
-  btime duration = last_time - loop_time;
-  if (duration > config->watchdog_warning)
+  btime duration = last_io_time - loop_time;
+  struct global_runtime *gr = atomic_load_explicit(&global_runtime, memory_order_relaxed);
+  if (duration > gr->watchdog_warning)
     log(L_WARN "I/O loop cycle took %u.%03u ms for %d events",
 	(uint) (duration TO_MS), (uint) (duration % 1000), event_log_num);
 }
@@ -2409,9 +2568,10 @@ watchdog_stop(void)
 void
 io_init(void)
 {
-  init_list(&sock_list);
-  init_list(&global_event_list);
-  init_list(&global_work_list);
+  init_list(&main_birdloop.sock_list);
+  ev_init_list(&global_event_list, &main_birdloop, "Global event list");
+  ev_init_list(&global_work_list, &main_birdloop, "Global work list");
+  ev_init_list(&main_birdloop.event_list, &main_birdloop, "Global fast event list");
   krt_io_init();
   // XXX init_times();
   // XXX update_times();
@@ -2425,64 +2585,46 @@ static int short_loops = 0;
 #define SHORT_LOOP_MAX 10
 #define WORK_EVENTS_MAX 10
 
+sock *stored_sock;
+
 void
 io_loop(void)
 {
   int poll_tout, timeout;
-  int nfds, events, pout;
+  int events, pout;
   timer *t;
-  sock *s;
-  node *n;
-  int fdmax = 256;
-  struct pollfd *pfd = xmalloc(fdmax * sizeof(struct pollfd));
+  struct pfd pfd;
+  BUFFER_INIT(pfd.pfd, &root_pool, 16);
+  BUFFER_INIT(pfd.loop, &root_pool, 16);
 
   watchdog_start1();
   for(;;)
     {
-      times_update(&main_timeloop);
+      times_update();
       ev_run_list(&global_event_list);
       ev_run_list_limited(&global_work_list, WORK_EVENTS_MAX);
-      timers_fire(&main_timeloop);
+      ev_run_list(&main_birdloop.event_list);
+      timers_fire(&main_birdloop.time, 1);
       io_close_event();
 
-      events = !EMPTY_LIST(global_event_list) || !EMPTY_LIST(global_work_list);
+      events =
+	!ev_list_empty(&global_event_list) ||
+	!ev_list_empty(&global_work_list) ||
+	!ev_list_empty(&main_birdloop.event_list);
+
       poll_tout = (events ? 0 : 3000); /* Time in milliseconds */
-      if (t = timers_first(&main_timeloop))
+      if (t = timers_first(&main_birdloop.time))
       {
-	times_update(&main_timeloop);
+	times_update();
 	timeout = (tm_remains(t) TO_MS) + 1;
 	poll_tout = MIN(poll_tout, timeout);
       }
 
-      nfds = 0;
-      WALK_LIST(n, sock_list)
-	{
-	  pfd[nfds] = (struct pollfd) { .fd = -1 }; /* everything other set to 0 by this */
-	  s = SKIP_BACK(sock, n, n);
-	  if (s->rx_hook)
-	    {
-	      pfd[nfds].fd = s->fd;
-	      pfd[nfds].events |= POLLIN;
-	    }
-	  if (s->tx_hook && s->ttx != s->tpos)
-	    {
-	      pfd[nfds].fd = s->fd;
-	      pfd[nfds].events |= POLLOUT;
-	    }
-	  if (pfd[nfds].fd != -1)
-	    {
-	      s->index = nfds;
-	      nfds++;
-	    }
-	  else
-	    s->index = -1;
+      BUFFER_FLUSH(pfd.pfd);
+      BUFFER_FLUSH(pfd.loop);
 
-	  if (nfds >= fdmax)
-	    {
-	      fdmax *= 2;
-	      pfd = xrealloc(pfd, fdmax * sizeof(struct pollfd));
-	    }
-	}
+      pipe_pollin(&main_birdloop.thread->wakeup, &pfd);
+      sockets_prepare(&main_birdloop, &pfd);
 
       /*
        * Yes, this is racy. But even if the signal comes before this test
@@ -2491,21 +2633,21 @@ io_loop(void)
 
       if (async_config_flag)
 	{
-	  io_log_event(async_config, NULL);
+	  io_log_event(async_config, NULL, DL_EVENTS);
 	  async_config();
 	  async_config_flag = 0;
 	  continue;
 	}
       if (async_dump_flag)
 	{
-	  io_log_event(async_dump, NULL);
+	  io_log_event(async_dump, NULL, DL_EVENTS);
 	  async_dump();
 	  async_dump_flag = 0;
 	  continue;
 	}
       if (async_shutdown_flag)
 	{
-	  io_log_event(async_shutdown, NULL);
+	  io_log_event(async_shutdown, NULL, DL_EVENTS);
 	  async_shutdown();
 	  async_shutdown_flag = 0;
 	  continue;
@@ -2513,103 +2655,108 @@ io_loop(void)
 
       /* And finally enter poll() to find active sockets */
       watchdog_stop();
-      pout = poll(pfd, nfds, poll_tout);
+      birdloop_leave(&main_birdloop);
+      pout = poll(pfd.pfd.data, pfd.pfd.used, poll_tout);
+      birdloop_enter(&main_birdloop);
       watchdog_start();
 
       if (pout < 0)
 	{
 	  if (errno == EINTR || errno == EAGAIN)
 	    continue;
-	  die("poll: %m");
+	  bug("poll: %m");
 	}
       if (pout)
 	{
-	  times_update(&main_timeloop);
+	  if (pfd.pfd.data[0].revents & POLLIN)
+	  {
+	    /* IO loop reload requested */
+	    pipe_drain(&main_birdloop.thread->wakeup);
+	    atomic_fetch_and_explicit(&main_birdloop.thread_transition, ~LTT_PING, memory_order_acq_rel);
+	    continue;
+	  }
+
+	  times_update();
 
 	  /* guaranteed to be non-empty */
-	  current_sock = SKIP_BACK(sock, n, HEAD(sock_list));
+	  main_birdloop.sock_active = SKIP_BACK(sock, n, HEAD(main_birdloop.sock_list));
 
-	  while (current_sock)
+	  while (main_birdloop.sock_active)
+	  {
+	    sock *s = main_birdloop.sock_active;
+	    if (s->index != -1)
 	    {
-	      sock *s = current_sock;
-	      if (s->index == -1)
-		{
-		  current_sock = sk_next(s);
-		  goto next;
-		}
-
 	      int e;
 	      int steps;
 
 	      steps = MAX_STEPS;
-	      if (s->fast_rx && (pfd[s->index].revents & POLLIN) && s->rx_hook)
+	      if (s->fast_rx && (pfd.pfd.data[s->index].revents & POLLIN) && s->rx_hook)
 		do
 		  {
 		    steps--;
-		    io_log_event(s->rx_hook, s->data);
-		    e = sk_read(s, pfd[s->index].revents);
-		    if (s != current_sock)
-		      goto next;
+		    io_log_event(s->rx_hook, s->data, DL_SOCKETS);
+		    e = sk_read(s, pfd.pfd.data[s->index].revents);
 		  }
-		while (e && s->rx_hook && steps);
+		while (e && (main_birdloop.sock_active == s) && s->rx_hook && steps);
+
+	      if (s != main_birdloop.sock_active)
+		continue;
 
 	      steps = MAX_STEPS;
-	      if (pfd[s->index].revents & POLLOUT)
+	      if (pfd.pfd.data[s->index].revents & POLLOUT)
 		do
 		  {
 		    steps--;
-		    io_log_event(s->tx_hook, s->data);
+		    io_log_event(s->tx_hook, s->data, DL_SOCKETS);
 		    e = sk_write(s);
-		    if (s != current_sock)
-		      goto next;
 		  }
-		while (e && steps);
+		while (e && (main_birdloop.sock_active == s) && steps);
 
-	      current_sock = sk_next(s);
-	    next: ;
+	      if (s != main_birdloop.sock_active)
+		continue;
 	    }
 
+	    main_birdloop.sock_active = sk_next(s);
+	  }
+
 	  short_loops++;
 	  if (events && (short_loops < SHORT_LOOP_MAX))
 	    continue;
 	  short_loops = 0;
 
 	  int count = 0;
-	  current_sock = stored_sock;
-	  if (current_sock == NULL)
-	    current_sock = SKIP_BACK(sock, n, HEAD(sock_list));
+	  main_birdloop.sock_active = stored_sock;
+	  if (main_birdloop.sock_active == NULL)
+	    main_birdloop.sock_active = SKIP_BACK(sock, n, HEAD(main_birdloop.sock_list));
 
-	  while (current_sock && count < MAX_RX_STEPS)
+	  while (main_birdloop.sock_active && count < MAX_RX_STEPS)
 	    {
-	      sock *s = current_sock;
+	      sock *s = main_birdloop.sock_active;
 	      if (s->index == -1)
-		{
-		  current_sock = sk_next(s);
-		  goto next2;
-		}
+		goto next2;
 
-	      if (!s->fast_rx && (pfd[s->index].revents & POLLIN) && s->rx_hook)
+	      if (!s->fast_rx && (pfd.pfd.data[s->index].revents & POLLIN) && s->rx_hook)
 		{
 		  count++;
-		  io_log_event(s->rx_hook, s->data);
-		  sk_read(s, pfd[s->index].revents);
-		  if (s != current_sock)
-		    goto next2;
+		  io_log_event(s->rx_hook, s->data, DL_SOCKETS);
+		  sk_read(s, pfd.pfd.data[s->index].revents);
+		  if (s != main_birdloop.sock_active)
+		    continue;
 		}
 
-	      if (pfd[s->index].revents & (POLLHUP | POLLERR))
+	      if (pfd.pfd.data[s->index].revents & (POLLHUP | POLLERR))
 		{
-		  sk_err(s, pfd[s->index].revents);
-		  if (s != current_sock)
-		    goto next2;
+		  sk_err(s, pfd.pfd.data[s->index].revents);
+		  if (s != main_birdloop.sock_active)
+		    continue;
 		}
 
-	      current_sock = sk_next(s);
 	    next2: ;
+	      main_birdloop.sock_active = sk_next(s);
 	    }
 
 
-	  stored_sock = current_sock;
+	  stored_sock = main_birdloop.sock_active;
 	}
     }
 }
