@@ -74,7 +74,7 @@ static list krt_proto_list;
 void
 krt_io_init(void)
 {
-  krt_pool = rp_new(&root_pool, "Kernel Syncer");
+  krt_pool = rp_new(&root_pool, the_bird_domain.the_bird, "Kernel Syncer");
   krt_filter_lp = lp_new_default(krt_pool);
   init_list(&krt_proto_list);
   krt_sys_io_init();
@@ -160,7 +160,16 @@ kif_shutdown(struct proto *P)
   kif_sys_shutdown(p);
   kif_proto = NULL;
 
-  return PS_DOWN;
+  return PS_FLUSH;
+}
+
+static void
+kif_cleanup(struct proto *p)
+{
+  if (p->debug & D_EVENTS)
+    log(L_TRACE "%s: Flushing interfaces", p->name);
+  if_start_update();
+  if_end_update();
 }
 
 static int
@@ -232,13 +241,14 @@ kif_copy_config(struct proto_config *dest, struct proto_config *src)
 struct protocol proto_unix_iface = {
   .name = 		"Device",
   .template = 		"device%d",
-  .class =		PROTOCOL_DEVICE,
   .proto_size =		sizeof(struct kif_proto),
   .config_size =	sizeof(struct kif_config),
+  .startup =		PROTOCOL_STARTUP_NECESSARY,
   .preconfig =		kif_preconfig,
   .init =		kif_init,
   .start =		kif_start,
   .shutdown =		kif_shutdown,
+  .cleanup =		kif_cleanup,
   .reconfigure =	kif_reconfigure,
   .copy_config =	kif_copy_config
 };
@@ -258,14 +268,14 @@ static inline void
 krt_trace_in(struct krt_proto *p, rte *e, char *msg)
 {
   if (p->p.debug & D_PACKETS)
-    log(L_TRACE "%s: %N: %s", p->p.name, e->net->n.addr, msg);
+    log(L_TRACE "%s: %N: %s", p->p.name, e->net, msg);
 }
 
 static inline void
 krt_trace_in_rl(struct tbf *f, struct krt_proto *p, rte *e, char *msg)
 {
   if (p->p.debug & D_PACKETS)
-    log_rl(f, L_TRACE "%s: %N: %s", p->p.name, e->net->n.addr, msg);
+    log_rl(f, L_TRACE "%s: %N: %s", p->p.name, e->net, msg);
 }
 
 /*
@@ -287,259 +297,40 @@ static struct tbf rl_alien = TBF_DEFAULT_LOG_LIMITS;
 static inline u32
 krt_metric(rte *a)
 {
-  eattr *ea = ea_find(a->attrs->eattrs, EA_KRT_METRIC);
+  eattr *ea = ea_find(a->attrs, &ea_krt_metric);
   return ea ? ea->u.data : 0;
 }
 
-static inline int
-krt_same_key(rte *a, rte *b)
-{
-  return (krt_metric(a) == krt_metric(b));
-}
-
-static inline int
-krt_uptodate(rte *a, rte *b)
-{
-  return (a->attrs == b->attrs);
-}
-
-static void
-krt_learn_announce_update(struct krt_proto *p, rte *e)
-{
-  net *n = e->net;
-  rta *aa = rta_clone(e->attrs);
-  rte *ee = rte_get_temp(aa, p->p.main_source);
-  rte_update(&p->p, n->n.addr, ee);
-}
-
-static void
-krt_learn_announce_delete(struct krt_proto *p, net *n)
-{
-  rte_update(&p->p, n->n.addr, NULL);
-}
-
 static void
 krt_learn_alien_attr(struct channel *c, rte *e)
 {
-  ASSERT(!e->attrs->cached);
-  e->attrs->pref = c->preference;
-
-  e->attrs = rta_lookup(e->attrs);
+  ea_set_attr_u32(&e->attrs, &ea_gen_preference, 0, c->preference);
 }
 
 /* Called when alien route is discovered during scan */
 static void
 krt_learn_scan(struct krt_proto *p, rte *e)
 {
-  net *n0 = e->net;
-  net *n = net_get(p->krt_table, n0->n.addr);
-  rte *m, **mm;
+  rte e0 = {
+    .attrs = e->attrs,
+    .src = rt_get_source(&p->p, krt_metric(e)),
+  };
 
-  krt_learn_alien_attr(p->p.main_channel, e);
+  krt_learn_alien_attr(p->p.main_channel, &e0);
 
-  for(mm=&n->routes; m = *mm; mm=&m->next)
-    if (krt_same_key(m, e))
-      break;
-  if (m)
-    {
-      if (krt_uptodate(m, e))
-	{
-	  krt_trace_in_rl(&rl_alien, p, e, "[alien] seen");
-	  rte_free(e);
-	  m->pflags |= KRT_REF_SEEN;
-	}
-      else
-	{
-	  krt_trace_in(p, e, "[alien] updated");
-	  *mm = m->next;
-	  rte_free(m);
-	  m = NULL;
-	}
-    }
-  else
-    krt_trace_in(p, e, "[alien] created");
-  if (!m)
-    {
-      e->next = n->routes;
-      n->routes = e;
-      e->pflags |= KRT_REF_SEEN;
-    }
-}
-
-static void
-krt_learn_prune(struct krt_proto *p)
-{
-  struct fib *fib = &p->krt_table->fib;
-  struct fib_iterator fit;
-
-  KRT_TRACE(p, D_EVENTS, "Pruning inherited routes");
-
-  FIB_ITERATE_INIT(&fit, fib);
-again:
-  FIB_ITERATE_START(fib, &fit, net, n)
-    {
-      rte *e, **ee, *best, **pbest, *old_best;
-
-      /*
-       * Note that old_best may be NULL even if there was an old best route in
-       * the previous step, because it might be replaced in krt_learn_scan().
-       * But in that case there is a new valid best route.
-       */
-
-      old_best = NULL;
-      best = NULL;
-      pbest = NULL;
-      ee = &n->routes;
-      while (e = *ee)
-	{
-	  if (e->pflags & KRT_REF_BEST)
-	    old_best = e;
-
-	  if (!(e->pflags & KRT_REF_SEEN))
-	    {
-	      *ee = e->next;
-	      rte_free(e);
-	      continue;
-	    }
-
-	  if (!best || krt_metric(best) > krt_metric(e))
-	    {
-	      best = e;
-	      pbest = ee;
-	    }
-
-	  e->pflags &= ~(KRT_REF_SEEN | KRT_REF_BEST);
-	  ee = &e->next;
-	}
-      if (!n->routes)
-	{
-	  DBG("%I/%d: deleting\n", n->n.prefix, n->n.pxlen);
-	  if (old_best)
-	    krt_learn_announce_delete(p, n);
-
-	  FIB_ITERATE_PUT(&fit);
-	  fib_delete(fib, n);
-	  goto again;
-	}
-
-      best->pflags |= KRT_REF_BEST;
-      *pbest = best->next;
-      best->next = n->routes;
-      n->routes = best;
-
-      if ((best != old_best) || p->reload)
-	{
-	  DBG("%I/%d: announcing (metric=%d)\n", n->n.prefix, n->n.pxlen, krt_metric(best));
-	  krt_learn_announce_update(p, best);
-	}
-      else
-	DBG("%I/%d: uptodate (metric=%d)\n", n->n.prefix, n->n.pxlen, krt_metric(best));
-    }
-  FIB_ITERATE_END;
-
-  p->reload = 0;
+  rte_update(p->p.main_channel, e->net, &e0, e0.src);
+  rt_unlock_source(e0.src);
 }
 
 static void
 krt_learn_async(struct krt_proto *p, rte *e, int new)
 {
-  net *n0 = e->net;
-  net *n = net_get(p->krt_table, n0->n.addr);
-  rte *g, **gg, *best, **bestp, *old_best;
-
-  krt_learn_alien_attr(p->p.main_channel, e);
-
-  old_best = n->routes;
-  for(gg=&n->routes; g = *gg; gg = &g->next)
-    if (krt_same_key(g, e))
-      break;
   if (new)
-    {
-      if (g)
-	{
-	  if (krt_uptodate(g, e))
-	    {
-	      krt_trace_in(p, e, "[alien async] same");
-	      rte_free(e);
-	      return;
-	    }
-	  krt_trace_in(p, e, "[alien async] updated");
-	  *gg = g->next;
-	  rte_free(g);
-	}
-      else
-	krt_trace_in(p, e, "[alien async] created");
+    return krt_learn_scan(p, e);
 
-      e->next = n->routes;
-      n->routes = e;
-    }
-  else if (!g)
-    {
-      krt_trace_in(p, e, "[alien async] delete failed");
-      rte_free(e);
-      return;
-    }
-  else
-    {
-      krt_trace_in(p, e, "[alien async] removed");
-      *gg = g->next;
-      rte_free(e);
-      rte_free(g);
-    }
-  best = n->routes;
-  bestp = &n->routes;
-  for(gg=&n->routes; g=*gg; gg=&g->next)
-  {
-    if (krt_metric(best) > krt_metric(g))
-      {
-	best = g;
-	bestp = gg;
-      }
-
-    g->pflags &= ~KRT_REF_BEST;
-  }
-
-  if (best)
-    {
-      best->pflags |= KRT_REF_BEST;
-      *bestp = best->next;
-      best->next = n->routes;
-      n->routes = best;
-    }
-
-  if (best != old_best)
-    {
-      DBG("krt_learn_async: distributing change\n");
-      if (best)
-	krt_learn_announce_update(p, best);
-      else
-	krt_learn_announce_delete(p, n);
-    }
-}
-
-static void
-krt_learn_init(struct krt_proto *p)
-{
-  if (KRT_CF->learn)
-  {
-    struct rtable_config *cf = mb_allocz(p->p.pool, sizeof(struct rtable_config));
-    cf->name = "Inherited";
-    cf->addr_type = p->p.net_type;
-    cf->internal = 1;
-
-    p->krt_table = rt_setup(p->p.pool, cf);
-  }
-}
-
-static void
-krt_dump(struct proto *P, struct dump_request *dreq)
-{
-  struct krt_proto *p = (struct krt_proto *) P;
-
-  if (!KRT_CF->learn)
-    return;
-  RDUMP("KRT: Table of inheritable routes\n");
-  rt_dump(dreq, p->krt_table);
+  struct rte_src *src = rt_get_source(&p->p, krt_metric(e));
+  rte_update(p->p.main_channel, e->net, NULL, src);
+  rt_unlock_source(src);
 }
 
 #endif
@@ -548,80 +339,20 @@ krt_dump(struct proto *P, struct dump_request *dreq)
  *	Routes
  */
 
-static inline int
-krt_is_installed(struct krt_proto *p, net *n)
-{
-  return n->routes && bmap_test(&p->p.main_channel->export_map, n->routes->id);
-}
+/* Hook defined in nest/rt-table.c ... to be refactored away later */
+rte *krt_export_net(struct channel *c, const net_addr *a, linpool *lp);
 
-static void
-krt_flush_routes(struct krt_proto *p)
-{
-  struct rtable *t = p->p.main_channel->table;
-
-  KRT_TRACE(p, D_EVENTS, "Flushing kernel routes");
-  FIB_WALK(&t->fib, net, n)
-    {
-      if (krt_is_installed(p, n))
-	{
-	  /* FIXME: this does not work if gw is changed in export filter */
-	  krt_replace_rte(p, n, NULL, n->routes);
-	}
-    }
-  FIB_WALK_END;
-}
-
-static struct rte *
-krt_export_net(struct krt_proto *p, net *net, rte **rt_free)
-{
-  struct channel *c = p->p.main_channel;
-  const struct filter *filter = c->out_filter;
-  rte *rt;
-
-  if (c->ra_mode == RA_MERGED)
-    return rt_export_merged(c, net, rt_free, krt_filter_lp, 1);
-
-  rt = net->routes;
-  *rt_free = NULL;
-
-  if (!rte_is_valid(rt))
-    return NULL;
-
-  if (filter == FILTER_REJECT)
-    return NULL;
-
-  /* We could run krt_preexport() here, but it is already handled by krt_is_installed() */
-
-  if (filter == FILTER_ACCEPT)
-    goto accept;
-
-  if (f_run(filter, &rt, krt_filter_lp, FF_SILENT) > F_ACCEPT)
-    goto reject;
-
-
-accept:
-  if (rt != net->routes)
-    *rt_free = rt;
-  return rt;
-
-reject:
-  if (rt != net->routes)
-    rte_free(rt);
-  return NULL;
-}
+static void krt_rt_notify(struct proto *P, struct channel *ch, const net_addr *net, rte *new, const rte *old);
 
 static int
 krt_same_dest(rte *k, rte *e)
 {
-  rta *ka = k->attrs, *ea = e->attrs;
+  ea_list *ka = k->attrs, *ea = e->attrs;
 
-  if (ka->dest != ea->dest)
-    return 0;
+  eattr *nhea_k = ea_find(ka, &ea_gen_nexthop);
+  eattr *nhea_e = ea_find(ea, &ea_gen_nexthop);
 
-  if (ka->dest == RTD_UNICAST)
-    return nexthop_equal(&(ka->nh), &(ea->nh));
-
-  return 1;
+  return (!nhea_k == !nhea_e) && adata_same(nhea_k->u.ptr, nhea_e->u.ptr);
 }
 
 /*
@@ -632,43 +363,43 @@ krt_same_dest(rte *k, rte *e)
 void
 krt_got_route(struct krt_proto *p, rte *e, s8 src)
 {
-  rte *new = NULL, *rt_free = NULL;
-  net *n = e->net;
+  /* If we happen to get an asynchronous route notification
+   * before initialization, we wait for the scan. */
+  if (p->sync_state == KPS_INIT)
+    return;
+
+  rte *new = NULL;
   e->pflags = 0;
 
 #ifdef KRT_ALLOW_LEARN
   switch (src)
     {
     case KRT_SRC_REDIRECT:
-      goto delete;
+      krt_trace_in(p, e, "deleting");
+      krt_replace_rte(p, e->net, NULL, e);
+      return;
 
     case KRT_SRC_KERNEL:
       if (KRT_CF->learn != KRT_LEARN_ALL)
-        goto ignore;
+      {
+	krt_trace_in(p, e, "ignored");
+	return;
+      }
       /* fallthrough */
 
-    case  KRT_SRC_ALIEN:
+    case KRT_SRC_ALIEN:
       if (KRT_CF->learn)
 	krt_learn_scan(p, e);
       else
-	{
-	  krt_trace_in_rl(&rl_alien, p, e, "[alien] ignored");
-	  rte_free(e);
-	}
+	krt_trace_in_rl(&rl_alien, p, e, "[alien] ignored");
       return;
     }
 #endif
+
   /* The rest is for KRT_SRC_BIRD (or KRT_SRC_UNKNOWN) */
 
-
-  /* We wait for the initial feed to have correct installed state */
-  if (!p->ready)
-    goto ignore;
-
-  if (!krt_is_installed(p, n))
-    goto delete;
-
-  new = krt_export_net(p, n, &rt_free);
+  /* Get the exported version */
+  new = krt_export_net(p->p.main_channel, e->net, krt_filter_lp);
 
   /* Rejected by filters */
   if (!new)
@@ -695,75 +426,80 @@ aseen:
   krt_trace_in(p, e, "already seen");
   goto done;
 
-ignore:
-  krt_trace_in(p, e, "ignored");
-  goto done;
-
 update:
   krt_trace_in(p, new, "updating");
-  krt_replace_rte(p, n, new, e);
+  krt_replace_rte(p, e->net, new, e);
   goto done;
 
 delete:
   krt_trace_in(p, e, "deleting");
-  krt_replace_rte(p, n, NULL, e);
+  krt_replace_rte(p, e->net, NULL, e);
   goto done;
 
-done:
-  rte_free(e);
-
-  if (rt_free)
-    rte_free(rt_free);
+done:;
 
   lp_flush(krt_filter_lp);
 }
 
-static void
+static bool
 krt_init_scan(struct krt_proto *p)
 {
-  bmap_reset(&p->seen_map, 1024);
+  switch (p->sync_state)
+  {
+    case KPS_INIT:
+      /* Allow exports now */
+      p->p.rt_notify = krt_rt_notify;
+      channel_start_export(p->p.main_channel);
+      rt_refresh_begin(&p->p.main_channel->in_req);
+      p->sync_state = KPS_FIRST_SCAN;
+      return 1;
+
+    case KPS_IDLE:
+      rt_refresh_begin(&p->p.main_channel->in_req);
+      bmap_reset(&p->seen_map, 1024);
+      p->sync_state = KPS_SCANNING;
+      return 1;
+
+    case KPS_FIRST_SCAN:
+    case KPS_SCANNING:
+      bug("Kernel scan double-init");
+
+    case KPS_PRUNING:
+      log(L_WARN "%s: Can't scan, still pruning", p->p.name);
+      return 0;
+
+  }
+
+  bug("Bad kernel sync state");
 }
 
 static void
 krt_prune(struct krt_proto *p)
 {
-  struct rtable *t = p->p.main_channel->table;
-
-  KRT_TRACE(p, D_EVENTS, "Pruning table %s", t->name);
-  FIB_WALK(&t->fib, net, n)
+  switch (p->sync_state)
   {
-    if (p->ready && krt_is_installed(p, n) && !bmap_test(&p->seen_map, n->routes->id))
-    {
-      rte *rt_free = NULL;
-      rte *new = krt_export_net(p, n, &rt_free);
+    case KPS_INIT:
+    case KPS_IDLE:
+      bug("Kernel scan prune without scan");
 
-      if (new)
-      {
-	krt_trace_in(p, new, "installing");
-	krt_replace_rte(p, n, new, NULL);
-      }
+    case KPS_SCANNING:
+      channel_request_full_refeed(p->p.main_channel);
+      /* fall through */
+    case KPS_FIRST_SCAN:
+      p->sync_state = KPS_PRUNING;
+      KRT_TRACE(p, D_EVENTS, "Pruning table %s", p->p.main_channel->table->name);
+      rt_refresh_end(&p->p.main_channel->in_req);
+      break;
 
-      if (rt_free)
-	rte_free(rt_free);
+    case KPS_PRUNING:
+      bug("Kernel scan double-prune");
 
-      lp_flush(krt_filter_lp);
-    }
   }
-  FIB_WALK_END;
-
-#ifdef KRT_ALLOW_LEARN
-  if (KRT_CF->learn)
-    krt_learn_prune(p);
-#endif
-
-  if (p->ready)
-    p->initialized = 1;
 }
 
 void
 krt_got_route_async(struct krt_proto *p, rte *e, int new, s8 src)
 {
-  net *net = e->net;
   e->pflags = 0;
 
   switch (src)
@@ -776,7 +512,7 @@ krt_got_route_async(struct krt_proto *p, rte *e, int new, s8 src)
       if (new)
 	{
 	  krt_trace_in(p, e, "[redirect] deleting");
-	  krt_replace_rte(p, net, NULL, e);
+	  krt_replace_rte(p, e->net, NULL, e);
 	}
       /* If !new, it is probably echo of our deletion */
       break;
@@ -795,7 +531,6 @@ krt_got_route_async(struct krt_proto *p, rte *e, int new, s8 src)
 	}
 #endif
     }
-  rte_free(e);
 }
 
 
@@ -825,7 +560,8 @@ krt_scan_all(timer *t UNUSED)
   krt_do_scan(NULL);
 
   WALK_LIST2(p, n, krt_proto_list, krt_node)
-    krt_prune(p);
+    if ((p->sync_state == KPS_SCANNING) || (p->sync_state == KPS_FIRST_SCAN))
+      krt_prune(p);
 }
 
 static void
@@ -874,7 +610,9 @@ krt_scan(timer *t)
   kif_force_scan();
 
   KRT_TRACE(p, D_EVENTS, "Scanning routing table");
-  krt_init_scan(p);
+  if (!krt_init_scan(p))
+    return;
+
   krt_do_scan(p);
   krt_prune(p);
 }
@@ -941,39 +679,68 @@ krt_assume_onlink(struct iface *iface, int ipv6)
 static int
 krt_preexport(struct channel *C, rte *e)
 {
-  // struct krt_proto *p = (struct krt_proto *) P;
-  if (e->src->proto == C->proto)
+  /* The export should not start before proper sync */
+  ASSERT_DIE(SKIP_BACK(struct krt_proto, p, C->proto)->sync_state != KPS_INIT);
+
+  if (e->src->owner == &C->proto->sources)
+#ifdef CONFIG_SINGLE_ROUTE
+    return 1;
+#else
     return -1;
+#endif
 
   if (!krt_capable(e))
+  {
+    if (C->debug & D_ROUTES)
+      log(L_TRACE "%s.%s: refusing incapable route for %N",
+	  C->proto->name, C->name, e->net);
     return -1;
+  }
 
   return 0;
 }
 
 static void
-krt_rt_notify(struct proto *P, struct channel *ch UNUSED, net *net,
-	      rte *new, rte *old)
+krt_rt_notify(struct proto *P, struct channel *ch, const net_addr *net,
+	      rte *new, const rte *old)
 {
   struct krt_proto *p = (struct krt_proto *) P;
 
-  if (config->shutdown)
-    return;
-
 #ifdef CONFIG_SINGLE_ROUTE
-  /*
-   * Implicit withdraw - when the imported kernel route becomes the best one,
-   * we know that the previous one exported to the kernel was already removed,
-   * but if we processed the update as usual, we would send withdraw to the
-   * kernel, which would remove the new imported route instead.
-   */
-  rte *best = net->routes;
-  if (!new && best && (best->src->proto == P))
+  /* Got the same route as we imported. Keep it, do nothing. */
+  if (new && new->src->owner == &P->sources)
     return;
 #endif
 
-  if (p->initialized)		/* Before first scan we don't touch the routes */
-    krt_replace_rte(p, net, new, old);
+  switch (p->sync_state)
+  {
+    case KPS_INIT:
+      bug("Routes in init state should have been rejected by preexport.");
+
+    case KPS_IDLE:
+    case KPS_PRUNING:
+      if (new && bmap_test(&p->seen_map, new->id))
+      {
+	if (ch->debug & D_ROUTES)
+	{
+	  /* Already installed and seen in the kernel dump */
+	  log(L_TRACE "%s.%s: %N already in kernel",
+	      P->name, ch->name, net);
+	}
+	return;
+      }
+
+      /* fall through */
+    case KPS_FIRST_SCAN:
+    case KPS_SCANNING:
+      /* Actually replace the route */
+      krt_replace_rte(p, net, new, old);
+      if (ch->debug & D_ROUTES)
+	log(L_TRACE "%s.%s: %N %s kernel",
+	    P->name, ch->name, net, old ? "replaced in" : "added to");
+      break;
+
+  }
 }
 
 static void
@@ -993,29 +760,60 @@ krt_if_notify(struct proto *P, uint flags, struct iface *iface UNUSED)
     krt_scan_timer_kick(p);
 }
 
-static void
-krt_reload_routes(struct channel *C)
+static int
+krt_reload_routes(struct channel *C, struct rt_feeding_request *rfr)
 {
   struct krt_proto *p = (void *) C->proto;
 
-  /* Although we keep learned routes in krt_table, we rather schedule a scan */
-
   if (KRT_CF->learn)
   {
-    p->reload = 1;
     krt_scan_timer_kick(p);
   }
+
+  if (rfr)
+    CALL(rfr->done, rfr);
+
+  return 1;
 }
 
+static void krt_cleanup(struct krt_proto *p);
+
 static void
-krt_feed_end(struct channel *C)
+krt_export_fed(struct channel *C)
 {
   struct krt_proto *p = (void *) C->proto;
 
-  p->ready = 1;
-  krt_scan_timer_kick(p);
+  switch (p->sync_state)
+  {
+    case KPS_INIT:
+      bug("KRT export started before scan");
+
+    case KPS_IDLE:
+      krt_scan_timer_kick(p);
+      break;
+
+    case KPS_FIRST_SCAN:
+      bug("KRT export done before first scan");
+
+    case KPS_SCANNING:
+      break;
+
+    case KPS_PRUNING:
+      KRT_TRACE(p, D_EVENTS, "Table %s pruned", p->p.main_channel->table->name);
+      p->sync_state = KPS_IDLE;
+      break;
+
+  }
 }
 
+static int
+krt_rte_better(const rte *new, const rte *old)
+{
+  u32 n = ea_get_int(new->attrs, &ea_krt_metric, IGP_METRIC_UNKNOWN);
+  u32 o = ea_get_int(old->attrs, &ea_krt_metric, IGP_METRIC_UNKNOWN);
+
+  return (n < o);
+}
 
 /*
  *	Protocol glue
@@ -1057,6 +855,10 @@ krt_postconfig(struct proto_config *CF)
   krt_sys_postconfig(cf);
 }
 
+struct rte_owner_class krt_rte_owner_class = {
+  .rte_better = krt_rte_better,
+};
+
 static struct proto *
 krt_init(struct proto_config *CF)
 {
@@ -1066,10 +868,13 @@ krt_init(struct proto_config *CF)
   p->p.main_channel = proto_add_channel(&p->p, proto_cf_main_channel(CF));
 
   p->p.preexport = krt_preexport;
-  p->p.rt_notify = krt_rt_notify;
-  p->p.if_notify = krt_if_notify;
+  /* Not setting rt_notify here to not start exports, must wait for the first scan
+   * and then we can start exports manually */
+  p->p.iface_sub.if_notify = krt_if_notify;
   p->p.reload_routes = krt_reload_routes;
-  p->p.feed_end = krt_feed_end;
+  p->p.export_fed = krt_export_fed;
+
+  p->p.sources.class = &krt_rte_owner_class;
 
   krt_sys_init(p);
   return &p->p;
@@ -1095,9 +900,9 @@ krt_start(struct proto *P)
   bmap_init(&p->seen_map, p->p.pool, 1024);
   add_tail(&krt_proto_list, &p->krt_node);
 
-#ifdef KRT_ALLOW_LEARN
-  krt_learn_init(p);
-#endif
+  /* Disable exports until initial scan is done.
+   * This is needed if somebody happens to restart the kernel protocol. */
+  p->p.rt_notify = NULL;
 
   if (!krt_sys_start(p))
   {
@@ -1120,21 +925,50 @@ krt_shutdown(struct proto *P)
 
   krt_scan_timer_stop(p);
 
-  /* FIXME we should flush routes even when persist during reconfiguration */
-  if (p->initialized && !KRT_CF->persist && (P->down_code != PDC_CMD_GR_DOWN))
-    krt_flush_routes(p);
-
-  p->ready = 0;
-  p->initialized = 0;
-
   if (p->p.proto_state == PS_START)
-    return PS_DOWN;
+    return PS_FLUSH;
+
+  /* FIXME we should flush routes even when persist during reconfiguration */
+  if ((p->sync_state != KPS_INIT) && !KRT_CF->persist && (P->down_code != PDC_CMD_GR_DOWN))
+  {
+    struct rt_export_feeder req = (struct rt_export_feeder)
+    {
+      .name = "shutdown.feeder",
+      .trace_routes = P->main_channel->debug,
+    };
+    rt_feeder_subscribe(&P->main_channel->table->export_best, &req);
+
+    RT_FEED_WALK(&req, f)
+    {
+      for (uint i = 0; i < f->count_routes; i++)
+      {
+	      rte *e = &f->block[i];
+        if (bmap_test(&P->main_channel->export_rejected_map, e->id))
+          continue;
+
+      /* if exported then delete from kernel */
+      krt_replace_rte(p, e->net, NULL, e);
+      }
+    }
+
+    krt_do_scan(p);
+    krt_cleanup(p);
+    proto_notify_state(&p->p, PS_FLUSH);
+    rt_feeder_unsubscribe(&req);
+  }
+  else
+    krt_cleanup(p);
+  return PS_FLUSH;
+}
+
+static void
+krt_cleanup(struct krt_proto *p)
+{
+  p->sync_state = KPS_INIT;
 
   krt_sys_shutdown(p);
   rem_node(&p->krt_node);
   bmap_free(&p->sync_map);
-
-  return PS_DOWN;
 }
 
 static int
@@ -1178,24 +1012,15 @@ krt_copy_config(struct proto_config *dest, struct proto_config *src)
   krt_sys_copy_config(d, s);
 }
 
-static int
-krt_get_attr(const eattr *a, byte *buf, int buflen)
-{
-  switch (a->id)
-  {
-  case EA_KRT_SOURCE:
-    bsprintf(buf, "source");
-    return GA_NAME;
-
-  case EA_KRT_METRIC:
-    bsprintf(buf, "metric");
-    return GA_NAME;
-
-  default:
-    return krt_sys_get_attr(a, buf, buflen);
-  }
-}
+struct ea_class ea_krt_source = {
+  .name = "krt_source",
+  .type = T_INT,
+};
 
+struct ea_class ea_krt_metric = {
+  .name = "krt_metric",
+  .type = T_INT,
+};
 
 #ifdef CONFIG_IP6_SADR_KERNEL
 #define MAYBE_IP6_SADR	NB_IP6_SADR
@@ -1212,11 +1037,11 @@ krt_get_attr(const eattr *a, byte *buf, int buflen)
 struct protocol proto_unix_kernel = {
   .name =		"Kernel",
   .template =		"kernel%d",
-  .class =		PROTOCOL_KERNEL,
   .preference =		DEF_PREF_INHERITED,
   .channel_mask =	NB_IP | MAYBE_IP6_SADR | MAYBE_MPLS,
   .proto_size =		sizeof(struct krt_proto),
   .config_size =	sizeof(struct krt_config),
+  .startup =		PROTOCOL_STARTUP_CONNECTOR,
   .preconfig =		krt_preconfig,
   .postconfig =		krt_postconfig,
   .init =		krt_init,
@@ -1224,14 +1049,15 @@ struct protocol proto_unix_kernel = {
   .shutdown =		krt_shutdown,
   .reconfigure =	krt_reconfigure,
   .copy_config =	krt_copy_config,
-  .get_attr =		krt_get_attr,
-#ifdef KRT_ALLOW_LEARN
-  .dump =		krt_dump,
-#endif
 };
 
 void
 krt_build(void)
 {
   proto_build(&proto_unix_kernel);
+
+  EA_REGISTER_ALL(
+      &ea_krt_source,
+      &ea_krt_metric,
+      );
 }
